#!/usr/bin/env python

# auto dj monitor
# Manuel Amador (Rudd-O)
# part of Smart DJ
# written on 2005-11-28

from ControllableThread import ControllableThread
from Queue import Queue,Empty
from threading import RLock,Thread
from amarokplus import DcopAmarokFailed, NotAnalyzedYet,UnAnalyzable
from Observable import Observer
import os
from songanalysiscommon import *
import logging
from sets import Set
import time

#remove dependency on Queue or make composition-based instead of inheritance-based
class SuperDynamicModeMonitor(ControllableThread,Queue,Observer):
		
	logger = logging.getLogger("SuperDynamicModeMonitor")
	
	def __init__(self,proxyObject,analysisProcess,configObject):
		ControllableThread.__init__(self)
		Queue.__init__(self)
		self.amarok = proxyObject
		self.analysisProcess = analysisProcess
		self.configObject = configObject
		self.sdm_enabled = False
		self.already_processed_files = Set()
		self.failed_cache = Set()
		self.songs_i_liked = []
		self.loop_running = RLock()
		self.observe(self.amarok) # we start observing amaroK
		self.observe(self.amarok.player) # the player
		self.observe(self.configObject) # and the configuration object
		self.amarok_available_yet = False
		
	def run(self):
		try: self._run()
		except DcopAmarokFailed,e:
			self.logger.debug("Aborting because amaroK is unexpectedly gone: %s",e.output)
		except:
			self.logger.exception("Something wrong in Auto DJ mode monitor, process stopped")
			say_sorry("Something went wrong in the Auto DJ monitor loop.  This is not your fault, but a bug in the Smart DJ plug-in.  Please report the contents of the file %s to the Smart DJ plug-in developers." %os.path.expanduser(n_logfile))
		
	def processEvent(self,notifierObject,eventName,*args):
		if eventName == "amaroKAvailable": self.amarok_available_yet = True
		elif eventName == "amaroKExit": self.kill(asynchronous=True) # we do it to get the fastest response here
		else: self.put(eventName) # the rest of events we delegate to a handler which can cope with queue overflow
	
	def _run(self):
		
		self.logger.debug("Auto DJ Monitor starting")
		while not self.amarok_available_yet and not self.killed(): time.sleep(1)
		self.logger.debug("Auto DJ Monitor started")
		
		self.on_start()
		while not self.killed():
			try: event = self.get(block=True,timeout=1)
			except Empty: continue
			self.logger.debug("Event received: %s",event)
			if event == "trackChange": self.on_track_changed()
			if event == "configChange": self.on_config_changed()
			if isinstance(event,Exception): raise event
			
		self.on_end()
		self.logger.debug("Auto DJ Monitor bailed")

	def on_start(self):
		self.on_config_changed()
		
		def run_loop_every_twennay():
			while True:
				time.sleep(20)
				self.run_super_dynamic_loop()
		
		t = Thread(target=run_loop_every_twennay)
		t.setDaemon(True)
		t.start()
	
	def on_end(self):
		self.logger.debug("About to enter on_end")
		self.restore_amarok_config()
	
	def on_track_changed(self):
		self.logger.debug("Auto DJ detected track change")
		self.run_super_dynamic_loop()
		
	def on_config_changed(self):
		self.logger.debug("Auto DJ Monitor detected config change or just started up, adjusting everything")
		
		if self.configObject["super_dynamic_mode"]["enabled"]:
			if not self.sdm_enabled:
				self.preserve_and_disable_amarok_config()
				self.logger.debug("Clearing already processed files and failed cache")
				self.already_processed_files = Set()
				self.failed_cache = Set()
				self.amarok.playlist.shortStatusMessage("Auto DJ mode is now <b>on</b>.  Please <b>do not enable</b> Random or Dynamic modes.")
				self.sdm_enabled = True
				# we just got enabled
			self.run_super_dynamic_loop() # this needs to run whether we're just enabled or not, to correct any "imperfections" and adjust to the users' requests
		else:
			if self.sdm_enabled:
				# we just got disabled
				self.sdm_enabled = False
				self.restore_amarok_config()
				self.amarok.playlist.shortStatusMessage("Auto DJ mode is now <b>off</b>.  Your Dynamic and Random settings have been restored.")
	
	def find_suitable_song(self,start_song,num_choices,spectrum_weight,tempo_weight,prune_songs,subquery=None):
		ce = self.amarok.collection
		
		num_results = num_choices + 1 + len(prune_songs) + 10 # the +1 is because we know the start song will come first, and the +10 is because we'll prune songs with the same file name!
		
		self.logger.debug("Finding %s songs tempo/spectrum %s/%s",num_results,tempo_weight,spectrum_weight)
		results = ce.getSimilarSongs(start_song,spectrum_weight,tempo_weight,num_results,subquery=subquery)
		results = [ r[0] for r in results ]
		
		if not results:
			self.logger.debug("Zero results from search, returning None")
			return None
		self.logger.debug("Search turned up %s raw results",len(results))

# 		prune_songs_cache = Set()
		if prune_songs:
			esequele = "select CONCAT( artist.name , ' - ' , tags.title ) from artist inner join tags on tags.artist = artist.id where tags.url in ( %s );" % " , ".join([ ce.sql_escape(e) for e in prune_songs ])
			prune_songs_cache = [ s.lower() for s in ce.query(esequele) ]
		else:
			prune_songs_cache = []
			
	# 	self.logger.debug("Results: %s\n",results)
		self.logger.debug("We will be pruning %s paths",len(prune_songs))
		self.logger.debug("We will also be pruning %s song titles",len(prune_songs_cache))
		
		nuevos_elementos = []
		for result in results:
			if result in prune_songs:
				self.logger.debug("Pruning %s from results list",result)
				continue
			
			try:
				artistalbumstring = ce.queryOne("select CONCAT( artist.name , ' - ' , tags.title ) from artist inner join tags on tags.artist = artist.id where tags.url = %s ;" % ce.sql_escape(result.lower()))
			except AttributeError: # this is to trap a bug someone reported
				self.logger.exception("Error!!!!: results: %s",results)
				raise
			
			if artistalbumstring in prune_songs_cache:
				self.logger.debug("Pruning %s from results list",artistalbumstring)
				continue
				
			nuevos_elementos.append(result)
			if len(nuevos_elementos) >= num_choices:
				self.logger.debug("Our pruned results list has achieved %s elements, breaking out of prune loop",num_choices)
				break
			
		results = nuevos_elementos
		if not results:
			self.logger.debug("Zero results after pruning, returning None")
			return None
		self.logger.debug("%s results after pruning",len(results))
		
		self.logger.debug("Randomly selecting among results")
		from random import choice
		return choice(results)


	def run_super_dynamic_loop(self): # Super Dynamic Loop!!!! TA-DAA!!!!!
		
		def run_if_unlocked():
			if not self.loop_running.acquire(blocking=0):
				return
			try:self._the_real_run()
			finally:self.loop_running.release()
		
		def run_guarded():
			try:
				run_if_unlocked()
			except DcopAmarokFailed,e:
				self.logger.debug("Aborting because amaroK is unexpectedly gone: %s",e.output)
			except Exception,e:
				self.logger.exception("Error while running super dynamic loop")
				self.processEvent(None,e)

		t = Thread(target=run_guarded)
		t.setDaemon(True)
		t.start()
		
	def _the_real_run(self):
		
		def count_upcoming_songs():
			active_index = self.amarok.playlist.getActiveIndex()
			if active_index is None: active_index = 0
			else: active_index = active_index + 1
			total_track_count = self.amarok.playlist.getTotalTrackCount()
			r = total_track_count - active_index
			self.logger.debug("active_index: %s    total track count: %s    r: %s",active_index,total_track_count,r)
			assert(r >= 0)
			return r
		
		sdm = self.configObject["super_dynamic_mode"]
		
		if sdm["enabled"]:
			playing_song = self.amarok.player.path()
			if not playing_song in self.songs_i_liked:
				# let's check whether we need to add this to the list of songs you liked
				playing_song_current_time = self.amarok.player.trackCurrentTime()
				playing_song_total_time = self.amarok.player.trackTotalTime() # in seconds
				if playing_song and playing_song_current_time and playing_song_total_time: # if there's a song playing
					playing_song_percent = float(playing_song_current_time) / float(playing_song_total_time)
					if playing_song_percent >= .40 or playing_song_current_time >= 60:
						self.logger.debug("We're liking %s, aren't we?"%playing_song)
						self.songs_i_liked.append(playing_song)
		
		now_upcoming_songs = count_upcoming_songs()
		if sdm["enabled"] and now_upcoming_songs < sdm["upcoming_songs"]:
			
			playlist_paths = [ a.Path for a in self.amarok.playlist.getTracks() ]
		
			# we now choose the start path
			if sdm["choice_mechanism"] in [SONG_IM_LISTENING,SONG_IM_LISTENING_WANDER]:
				current_url = self.amarok.player.path()
			elif sdm["choice_mechanism"] == LAST_SONG_IN_PLAYLIST:
				if len(playlist_paths): current_url = playlist_paths[-1]
				else: current_url = None
			elif sdm["choice_mechanism"] == LAST_SONG_I_LIKED:
				# if the user skipped once, count_upcoming_songs() DEBE ser por fuerza 1 menos que sdm["upcoming_songs"], pero si el usuario ha saltado MAS de una cancion, entonces de ley esta condicion no se cumple
				if  now_upcoming_songs == sdm["upcoming_songs"] - 1 and now_upcoming_songs > 0:
					# hang on, unless the current url is the song he liked
					if playing_song in self.songs_i_liked:
						self.logger.debug("Upcoming song count is not critically low, but the user already decided this song he liked, so we're using it as search source")
						current_url = playing_song
					else:
						self.logger.debug("Upcoming song count is not critically low, forfeiting this chance")
						current_url = None
				else:
					if self.songs_i_liked:
						self.logger.debug("Upcoming song count is CRITICALLY low (or the user skipped more than once), using the last liked song as search start instead of hanging on waiting and seeing whether the user likes the current song")
						current_url = self.songs_i_liked[-1]
					else:
						self.logger.debug("Upcoming song count is CRITICALLY low, BUT there is no last liked song") # I wonder, should we fall back to currently being played URI?
						current_url = None
			else: assert(False)
			
			if not current_url:
				self.logger.debug("No start song, bailing")
				return # nothing's playing so we cannot determine what to use as criteria for start_url
			
			# we initialize the exceptions set with the already processed files list
			paths_to_prune = Set(self.already_processed_files)
			self.logger.debug("Initialized paths to prune with %s paths",len(paths_to_prune))
			paths_to_prune.union_update([ p for p in playlist_paths if p ])
			self.logger.debug("After adding paths in playlist, we're on %s paths",len(paths_to_prune))
			
			def get_subquery():
				if not sdm["source_playlist"]: return None
				self.logger.debug("Using a subquery")
				try: subquery = self.amarok.smartPlaylistCollection[sdm["source_playlist"]].sqlquery
				except KeyError, key:
					self.logger.warning("Smart playlist %s not found, resetting",key)
					sdm["source_playlist"] = None
					subquery = None
				return subquery
			
			source_playlist = sdm["source_playlist"]
			subquery = get_subquery()
			
			#FIXME this is for debuging
			start_time_2 = time.time()
			while sdm["enabled"] and count_upcoming_songs() < sdm["upcoming_songs"]:
				
				start_time = time.time()
				current_song = self.amarok.collection.getSong(current_url)
				if source_playlist != sdm["source_playlist"]: # we need to reload the subquery
					source_playlist = sdm["source_playlist"] # remember the new source playlist
					subquery = get_subquery() # reset the subquery!

				self.logger.debug("Trying to find a similar song to %s",current_url)
				try:
					
# 					we add the path to the permanent and this run's list
					self.already_processed_files.add(current_url)
					paths_to_prune.add(current_url)
					
					if current_url not in self.failed_cache: # restrict showing of messages to tracks which did not fail in previous ocassions
						if sdm["top_choices"] == 1:
							self.amarok.playlist.shortStatusMessage("Auto DJ mode is looking for a track similar to %s"%(current_song.getLongName(True)))
						else:
							self.amarok.playlist.shortStatusMessage("Auto DJ mode is looking for %s tracks similar to %s"%(sdm["top_choices"],current_song.getLongName(True)))
					
					choice = self.find_suitable_song(
						current_url,sdm["top_choices"],
						sdm["spectrum_weight"],sdm["tempo_weight"],
						paths_to_prune,subquery=subquery)
					if choice:
						self.logger.debug("Choice: %s",choice)
						self.amarok.playlist.addMedia(choice)
						choice_song = self.amarok.collection.getSong(choice)
# 						we add the path to the permanent and this run's list
						self.already_processed_files.add(choice)
						paths_to_prune.add(choice)
						self.amarok.playlist.shortStatusMessage("Auto DJ mode added %s to the playlist"%(choice_song.getLongName(True)))
						if sdm["choice_mechanism"] in [SONG_IM_LISTENING_WANDER,
							LAST_SONG_IN_PLAYLIST]:
								current_url = choice # yes, sir!
					else: raise NoChoicesFound
				except NotAnalyzedYet:
					if current_url not in self.failed_cache: # restrict showing of messages to tracks which did not fail in previous ocassions
						self.logger.warning("%s was not analyzed yet, adding to priority queue",current_url)
						self.analysisProcess.addToPriorityQueue(current_url)
						self.amarok.playlist.shortStatusMessage("Auto DJ mode is analyzing %s to find similar tracks" 
							%(current_song.getLongName(True)))
						self.failed_cache.add(current_url)
					break
				except UnAnalyzable:
					self.logger.warning("%s could not be analyzed by songanalysis, breaking",current_url)
					if current_url not in self.failed_cache: # restrict showing of messages to tracks which did not fail in previous ocassions
# 						self.collection_expert.add_to_priority_queue(current_url)
						self.amarok.playlist.shortStatusMessage("Auto DJ mode <b>was unable</b> to analyze " %(current_song.getLongName(True)))
						self.failed_cache.add(current_url)
					break
				except NoChoicesFound:
					self.logger.warning("%s did not turn up any choices, breaking loop",current_url)
					if current_url not in self.failed_cache:
						self.amarok.playlist.shortStatusMessage("Auto DJ mode <b>could not find</b> any tracks similar to %s"%(current_song.getLongName(True)))
						self.failed_cache.add(current_url)
					break
				self.logger.debug("Loop took %s seconds",time.time() -start_time)
			self.logger.debug("Total run took %s seconds",time.time() -start_time_2)
	
	def preserve_and_disable_amarok_config(self):
		m = self
		m.random_mode_status = self.amarok.player.randomModeStatus()
		m.dynamic_mode_status = self.amarok.player.dynamicModeStatus()
		self.amarok.player.enableRandomMode(False)
		self.amarok.player.enableDynamicMode(False)
		
	def restore_amarok_config(self):
		m = self
		if hasattr(m,"random_mode_status"):
			self.amarok.player.enableRandomMode(m.random_mode_status)
		if hasattr(m,"dynamic_mode_status"):
			self.amarok.player.enableDynamicMode(m.dynamic_mode_status)
	

